@@ -4,7 +4,7 @@
  * Copyright (c) 2000-2006 Anton Altaparmakov
  * Copyright (c) 2002-2006 Szabolcs Szakacsits
  * Copyright (c) 2007      Yura Pakhuchiy
- * Copyright (c) 2011      Jean-Pierre Andre
+ * Copyright (c) 2011-2015 Jean-Pierre Andre
  *
  * This utility fixes some common NTFS problems, resets the NTFS journal file
  * and schedules an NTFS consistency check for the first boot into Windows.
@@ -123,7 +123,7 @@ struct MFT_SELF_LOCATED {
  * usage
  */
 __attribute__((noreturn))
-static void usage(void)
+static void usage(int ret)
 {
 	ntfs_log_info("%s v%s (libntfs-3g)\n"
 		   "\n"
@@ -140,7 +140,7 @@ static void usage(void)
 		   EXEC_NAME, VERSION, EXEC_NAME,
 		   EXEC_NAME);
 	ntfs_log_info("%s%s", ntfs_bugs, ntfs_home);
-	exit(1);
+	exit(ret);
 }
 
 /**
@@ -153,11 +153,11 @@ static void version(void)
 		   "Attempt to fix an NTFS partition.\n\n"
 		   "Copyright (c) 2000-2006 Anton Altaparmakov\n"
 		   "Copyright (c) 2002-2006 Szabolcs Szakacsits\n"
-		   "Copyright (c) 2007      Yura Pakhuchiy\n\n"
-		   "Copyright (c) 2011      Jean-Pierre Andre\n\n",
+		   "Copyright (c) 2007      Yura Pakhuchiy\n"
+		   "Copyright (c) 2011-2015 Jean-Pierre Andre\n\n",
 		   EXEC_NAME, VERSION);
 	ntfs_log_info("%s\n%s%s", ntfs_gpl, ntfs_bugs, ntfs_home);
-	exit(1);
+	exit(0);
 }
 
 /**
@@ -185,7 +185,7 @@ static void parse_options(int argc, char **argv)
 				opt.volume = argv[optind - 1];
 			else {
 				ntfs_log_info("ERROR: Too many arguments.\n");
-				usage();
+				usage(1);
 			}
 			break;
 		case 'b':
@@ -198,20 +198,21 @@ static void parse_options(int argc, char **argv)
 			opt.no_action = TRUE;
 			break;
 		case 'h':
+			usage(0);
 		case '?':
-			usage();
+			usage(1);
 			/* fall through */
 		case 'V':
 			version();
 		default:
 			ntfs_log_info("ERROR: Unknown option '%s'.\n", argv[optind - 1]);
-			usage();
+			usage(1);
 		}
 	}
 
 	if (opt.volume == NULL) {
 		ntfs_log_info("ERROR: You must specify a device.\n");
-		usage();
+		usage(1);
 	}
 }
 
@@ -1273,6 +1274,104 @@ static int try_alternate_boot(ntfs_volume *vol, char *full_bs,
 	return (res);
 }
 
+/*
+ *		Check and fix the alternate boot sector
+ *
+ *	The alternate boot sector is usually in the last sector of a
+ *	partition, which should not be used by the file system
+ *	(the sector count in the boot sector should be less than
+ *	the total sector count in the partition).
+ *
+ *	chkdsk never changes the count in the boot sector.
+ *	- If this is less than the total count, chkdsk place the
+ *	  alternate boot sector into the sector,
+ *	- if the count is the same as the total count, chkdsk place
+ *	  the alternate boot sector into the middle sector (half
+ *	  the total count rounded upwards)
+ *	- if the count is greater than the total count, chkdsk
+ *	  declares the file system as raw, and refuses to fix anything.
+ *
+ *	Here, we check and fix the alternate boot sector, only in the
+ *	first situation where the file system does not overflow on the
+ *	last sector.
+ *
+ *	Note : when shrinking a partition, ntfsresize cannot determine
+ *	the future size of the partition. As a consequence the number of
+ *	sectors in the boot sectors may be less than the possible size.
+ *
+ *	Returns 0 if successful
+ */
+
+static int check_alternate_boot(ntfs_volume *vol)
+{
+	s64 got_sectors;
+	s64 actual_sectors;
+	s64 last_sector_off;
+	char *full_bs;
+	char *alt_bs;
+	NTFS_BOOT_SECTOR *bs;
+	s64 br;
+	s64 bw;
+	int res;
+
+	res = -1;
+	full_bs = (char*)malloc(vol->sector_size);
+	alt_bs = (char*)malloc(vol->sector_size);
+	if (!full_bs || !alt_bs) {
+		ntfs_log_info("Error : failed to allocate memory\n");
+		goto error_exit;
+	}
+	/* Now read both bootsectors. */
+	br = ntfs_pread(vol->dev, 0, vol->sector_size, full_bs);
+	if (br == vol->sector_size) {
+		bs = (NTFS_BOOT_SECTOR*)full_bs;
+		got_sectors = le64_to_cpu(bs->number_of_sectors);
+		actual_sectors = ntfs_device_size_get(vol->dev,
+						vol->sector_size);
+		if (actual_sectors > got_sectors) {
+			last_sector_off = (actual_sectors - 1)
+						<< vol->sector_size_bits;
+			ntfs_log_info("Checking the alternate boot sector... ");
+			br = ntfs_pread(vol->dev, last_sector_off,
+						vol->sector_size, alt_bs);
+		} else {
+			ntfs_log_info("Checking file system overflow... ");
+			br = -1;
+		}
+		/* accept getting no byte, needed for short image files */
+		if (br >= 0) {
+			if ((br != vol->sector_size)
+			    || memcmp(full_bs, alt_bs, vol->sector_size)) {
+				if (opt.no_action) {
+					ntfs_log_info("BAD\n");
+				} else {
+					bw = ntfs_pwrite(vol->dev,
+						last_sector_off,
+						vol->sector_size, full_bs);
+					if (bw == vol->sector_size) {
+						ntfs_log_info("FIXED\n");
+						res = 0;
+					} else {
+						ntfs_log_info(FAILED);
+					}
+				}
+			} else {
+				ntfs_log_info(OK);
+				res = 0;
+			}
+		} else {
+			ntfs_log_info(FAILED);
+		}
+	} else {
+		ntfs_log_info("Error : could not read the boot sector again\n");
+	}
+	free(full_bs);
+	free(alt_bs);
+
+error_exit :
+	return (res);
+}
+
 /*
  *		Try to fix problems which may arise in the start up sequence
  *
@@ -1321,7 +1420,7 @@ static int fix_startup(struct ntfs_device *dev, unsigned long flags)
 	NVolSetShowSysFiles(vol);
 	NVolSetShowHidFiles(vol);
 	NVolClearHideDotFiles(vol);
-	if (flags & MS_RDONLY)
+	if (flags & NTFS_MNT_RDONLY)
 		NVolSetReadOnly(vol);
 	
 	/* ...->open needs bracketing to compile with glibc 2.7 */
@@ -1410,7 +1509,7 @@ static int fix_mount(void)
 		ntfs_log_perror("Failed to allocate device");
 		return -1;
 	}
-	flags = (opt.no_action ? MS_RDONLY : 0);
+	flags = (opt.no_action ? NTFS_MNT_RDONLY : 0);
 	vol = ntfs_volume_startup(dev, flags);
 	if (!vol) {
 		ntfs_log_info(FAILED);
@@ -1475,7 +1574,7 @@ int main(int argc, char **argv)
 		ntfs_log_perror("Failed to determine whether %s is mounted",
 				opt.volume);
 	/* Attempt a full mount first. */
-	flags = (opt.no_action ? MS_RDONLY : 0);
+	flags = (opt.no_action ? NTFS_MNT_RDONLY : 0);
 	ntfs_log_info("Mounting volume... ");
 	vol = ntfs_mount(opt.volume, flags);
 	if (vol) {
@@ -1495,6 +1594,10 @@ int main(int argc, char **argv)
 			exit(1);
 		}
 	}
+	if (check_alternate_boot(vol)) {
+		ntfs_log_error("Error: Failed to fix the alternate boot sector\n");
+		exit(1);
+	}
 	/* So the unmount does not clear it again. */
 
 	/* Porting note: The WasDirty flag was set here to prevent ntfs_unmount
@@ -1543,8 +1646,10 @@ int main(int argc, char **argv)
 	/* Set return code to 0. */
 	ret = 0;
 error_exit:
-	if (ntfs_umount(vol, 0))
-		ntfs_umount(vol, 1);
+	if (ntfs_umount(vol, 1)) {
+		ntfs_log_info("Failed to unmount partition\n");
+		ret = 1;
+	}
 	if (ret)
 		exit(ret);
 	return ret;
