@@ -56,8 +56,13 @@
 #else
 #define USESTUBS 0 /* direct calls to API, based on following definitions */
 #define ENVNTFS3G "NTFS3G"
-#define LIBFILE64 "/lib64/libntfs-3g.so.831"
-#define LIBFILE "/lib/libntfs-3g.so.831"
+#if defined(__SVR4)
+#define LIBFILE64 "/usr/lib/amd64/libntfs-3g.so.861"
+#define LIBFILE "/usr/lib/libntfs-3g.so.861"
+#else
+#define LIBFILE64 "/lib64/libntfs-3g.so.861"
+#define LIBFILE "/lib/libntfs-3g.so.861"
+#endif
 #endif
 
 #define MAPDIR ".NTFS-3G"
@@ -154,8 +159,8 @@ typedef int BOOL;   /* Already defined in windows.h */
 typedef u32 DWORD; /* must be 32 bits whatever the platform */
 typedef DWORD *LPDWORD;
 
-#define MS_NONE 0    /* no flag for mounting the device */
-#define MS_RDONLY 1  /* flag for mounting the device read-only */
+#define NTFS_MNT_NONE 0    /* no flag for mounting the device */
+#define NTFS_MNT_RDONLY 1  /* flag for mounting the device read-only */
 
 #endif /* WIN32 */
 
@@ -221,14 +226,6 @@ typedef DWORD *LPDWORD;
 
 typedef le16 ntfschar;
 
-typedef struct {
-	le32 a;
-	le16 b,c;
-	struct {
-		le16 m,n,o,p,  q,r,s,t;
-	} ;
-} GUID;
-
 #define ntfs_log_error(args...) do { printf("** " args); if (!isatty(1)) fprintf(stderr,args); } while(0)
 
 /*
@@ -524,7 +521,8 @@ enum {
 #define CONTAINER_INHERIT_ACE		  (0x2)
 #define NO_PROPAGATE_INHERIT_ACE	  (0x4)
 #define INHERIT_ONLY_ACE		  (0x8)
-#define VALID_INHERIT_FLAGS		  (0xF)
+#define INHERITED_ACE			  (0x10)
+#define VALID_INHERIT_FLAGS		  (0x1F)
 
 /*
  *	     Other useful definitions
@@ -541,6 +539,12 @@ enum {
 #define ACL_REVISION_DS 4
 #endif
 
+#ifndef INHERITED_ACE /* not always defined in <windows.h> */
+#define INHERITED_ACE			  (0x10)
+#undef VALID_INHERIT_FLAGS
+#define VALID_INHERIT_FLAGS		  (0x1F)
+#endif
+
 /*
  *		Matching of ntfs permissions to Linux permissions
  *	these constants are adapted to endianness
@@ -562,12 +566,14 @@ enum {
           /* flags tested for meaning exec, write or read */
 	  /* tests for write allow for interpretation of a sticky bit */
 
-#define FILE_GREAD (FILE_READ_DATA | GENERIC_READ)
-#define FILE_GWRITE (FILE_WRITE_DATA | FILE_APPEND_DATA | GENERIC_WRITE)
-#define FILE_GEXEC (FILE_EXECUTE | GENERIC_EXECUTE)
-#define DIR_GREAD (FILE_LIST_DIRECTORY | GENERIC_READ)
-#define DIR_GWRITE (FILE_ADD_FILE | FILE_ADD_SUBDIRECTORY | GENERIC_WRITE)
-#define DIR_GEXEC (FILE_TRAVERSE | GENERIC_EXECUTE)
+#define FILE_GREAD (FILE_READ_DATA | GENERIC_READ | GENERIC_ALL)
+#define FILE_GWRITE (FILE_WRITE_DATA | FILE_APPEND_DATA | GENERIC_WRITE \
+			| GENERIC_ALL)
+#define FILE_GEXEC (FILE_EXECUTE | GENERIC_EXECUTE | GENERIC_ALL)
+#define DIR_GREAD (FILE_LIST_DIRECTORY | GENERIC_READ | GENERIC_ALL)
+#define DIR_GWRITE (FILE_ADD_FILE | FILE_ADD_SUBDIRECTORY | GENERIC_WRITE \
+			| GENERIC_ALL)
+#define DIR_GEXEC (FILE_TRAVERSE | GENERIC_EXECUTE | GENERIC_ALL)
 
 	/* standard owner (and administrator) rights */
 
